"""
Author: Lukas Runt
Email: lrunt@students.zcu.cz
Date: November 04, 2023
Version: 1.0.0
Description: Inference engine of expert system, communicates with all other modules and generating the result
"""

import KomunikacniModul
import VysvetlovaciModul
import BazeZnalosti
import BazeDat

import sys

# Type of input what we want
INTEGER_TYPE = "INT"


def app():
    """
    The main method which controls everything
    """
    print(KomunikacniModul.welcome_text)
    rooms = get_parameters()
    lights = []
    required_lumen_of_rooms = []
    recommended_colors = []
    for i in range(len(rooms)):
        # Adding basic lights
        lights_in_the_room = {}
        required_lumen = calculate_lumen(rooms[i].room_type, rooms[i].area)
        required_lumen_of_rooms.append(required_lumen)
        recommended_colors.append(BazeZnalosti.recommended_colors[rooms[i].room_type])
        color_interval = BazeZnalosti.colors[BazeZnalosti.recommended_colors[rooms[i].room_type]]
        basic_lights = filter_lights_by_speciality(BazeDat.lights, "BASIC")
        filtered_lights = filter_lights_by_required_luminosity(basic_lights, color_interval)
        lights_in_the_room["BASIC"] = get_count_of_lights(select_lights(filtered_lights, required_lumen))
        # Adding special lights
        for speciality in rooms[i].speciality:
            add_special_lights(BazeDat.lights, lights_in_the_room, speciality, color_interval,
                               rooms[i].speciality[speciality])
        lights.append(lights_in_the_room)
    VysvetlovaciModul.print_result(lights, rooms, required_lumen_of_rooms, recommended_colors)


def get_integer_input(minimum, maximum):
    """
    Method controls correctness of user input
    Controls if input is integer, if yes then controls if integer is in interval
    :param minimum: minimal possible number
    :param maximum: maximal possible number
    :return: -1 = Wrong input, or number
    """
    user_input = input()
    try:
        integer_value = int(user_input)
        if minimum <= integer_value <= maximum:
            return integer_value
        else:
            print(KomunikacniModul.invalid_input)
    except ValueError:
        print(KomunikacniModul.invalid_input)
    return -1


def get_input(text, input_type=None, minimum=0, maximum=sys.maxsize):
    """
    Method gets input from the user and controls its correctness
    If input is not correct, then method prints text again and user must enter input again
    :param text: instructions to the user
    :param input_type: type of input, which data type we want (optional)
    :param minimum: minimum of the interval (optional)
    :param maximum: maximum of the interval (optional)
    :return: user input
    """
    user_input = -1
    if input_type == INTEGER_TYPE:
        while user_input == -1:
            print(text, end="")
            user_input = get_integer_input(minimum, maximum)
    else:
        print(text, end="")
        user_input = input()
    return user_input


def get_parameters():
    """
    Method gets all important information about rooms
    :return: All information about rooms
    """
    number_of_rooms = get_input(KomunikacniModul.number_of_rooms, INTEGER_TYPE)
    rooms = []
    for i in range(number_of_rooms):
        print(KomunikacniModul.separator)
        rooms.append(get_room_parameters())
    return rooms


def get_room_parameters():
    """
    Method to get all important information about one room
    :return: all information about one room
    """
    print(KomunikacniModul.list_of_rooms)
    VysvetlovaciModul.print_options(BazeDat.room_types)
    room_type = BazeDat.room_types[get_input("Vyber místnost: ", INTEGER_TYPE, 0, len(BazeDat.room_types)) - 1]
    room_area = get_input(KomunikacniModul.enter_room_area, INTEGER_TYPE)
    speciality = get_speciality(room_type)
    return BazeDat.Room(room_type, room_area, speciality)


def get_speciality(room_type):
    """
    Method gets if the user want to have special lights in the room
    :param room_type: type of the room
    :return: if user wants specialities, and how many - dictionary {key = speciality, value = number of  special lights}
    """
    specialities = BazeZnalosti.special_option[room_type]
    user_choice = {}
    for i in range(len(specialities)):
        if specialities[i] != "LINKA":
            if specialities[i] == "DECORATION":
                user_input = get_input(KomunikacniModul.special_option[specialities[i]], INTEGER_TYPE, 0, 1)
                if user_input != 0:
                    user_choice[specialities[i]] = user_input
            else:
                user_input = get_input(KomunikacniModul.special_option[specialities[i]], INTEGER_TYPE)
                if user_input != 0:
                    user_choice[specialities[i]] = user_input
        else:
            user_choice[specialities[i]] = 1
    return user_choice


def calculate_lumen(room_type, room_area):
    """
    Method calculates how many lumen we need in the room
    :param room_type: type of the room
    :param room_area: area of the room in square meters [m^2]
    :return: required lumen in the room
    """
    return BazeZnalosti.required_lux[room_type] * room_area


def filter_lights_by_required_luminosity(lights, color_range):
    """
    Method filter lights by color range in Kevin [K]
    :param lights: all lights to filter
    :param color_range: interval of color, in what the light must fit
    :return: filtered lights
    """
    return [light for light in lights if color_range[0] <= light.color <= color_range[1]]


def filter_lights_by_speciality(lights, speciality):
    """
    Method filter lights by its speciality
    :param lights: all lights to filter
    :param speciality: speciality which the color must have
    :return: filtered lights
    """
    return [light for light in lights if BazeZnalosti.speciality_of_type[light.light_type] == speciality]


def select_lights(lights, lumen):
    """
    Method selecting the right lights
    :param lights: All lights what can be used
    :param lumen: number lumen the lights must have
    :return: picked lights
    """
    selected_lights = []
    # sorting_lights_by_suitability
    while lumen > 0:
        sorted_objects = sorted(lights, key=lambda obj: abs(obj.luminosity - lumen))
        selected_light = sorted_objects[0]
        lumen -= selected_light.luminosity
        selected_lights.append(selected_light)
    return selected_lights


def add_special_lights(lights, selected_lights, speciality, color_range, number):
    """
    Method adds special lights what user wants
    :param lights: all possible lights
    :param selected_lights: selected lights from to light the room
    :param speciality: speciality what user wants
    :param color_range: color range what is good the light be
    :param number: number of light, what user wants
    """
    lights_with_speciality = filter_lights_by_speciality(lights, speciality)
    lights_with_right_color = filter_lights_by_required_luminosity(lights_with_speciality, color_range)
    if len(lights_with_right_color) != 0:
        special_lights = {lights_with_right_color[0].name: number}
        selected_lights[speciality] = special_lights
    else:
        special_lights = {lights_with_speciality[0].name: number}
        selected_lights[speciality] = special_lights


def get_count_of_lights(lights_for_the_room):
    """
    Method converts lights to right format for explanation module
    :param lights_for_the_room: is lights what was chosen for the room
    :return: right format of lights
    """
    lights = {}
    for light in lights_for_the_room:
        if light.name not in lights:
            lights[light.name] = 1
        else:
            lights[light.name] = lights[light.name] + 1
    return lights
